import os.path


# general constants
HOME = os.path.expanduser("~")
PYRUSTIC_HOME = os.path.join(HOME, "pyrustic")
JESTH_DIR = os.path.join(PYRUSTIC_HOME, "jesth")  # free for storing random jesthfiles
INDENT_WIDTH = 4  # Four spaces
